/*
* This file is part of Firestorm NIDS
* Copyright (c) 2002 Gianni Tedesco
* This program is released under the terms of the GNU GPL version 2
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <sys/uio.h>

#include <firestorm.h>
#include <stormwall.h>
#include <packet.h>
#include <generator.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <cleanup.h>
#include <sensor.h>
#include <capture.h>
#include <capdev.h>
#include <args.h>
#include <target.h>
#include <elog.h>

/* Current alert log file */
extern char *fn_alert;
int log_fd=-1;

/* stormwall FIFO */
static char fifo_fn[]="./log/.stormwall";
int fifo_fd=-1;

/* stormwall mode */
#define STORMWALL_NONE 0 /* stormwall comms disabled */
#define STORMWALL_WAIT 1 /* wait for stormwall to come alive */
#define STORMWALL_FAIL 2 /* fail immidietly if stormwall is not ready */
unsigned int stormwall_mode=STORMWALL_NONE;

/* Rotation stuff (filesize) */
unsigned int max_bytes=(256*1024); /* 256KB default */
unsigned int last_bytes=sizeof(struct efile_hdr);

/* Rotation stuff (time limit) */
unsigned int max_time=60; /* 1hr default */
unsigned int last_time=0;

/* Values used for rotation filenames */
unsigned int first_packet=1;
struct timeval rotate_time;

void alert_rotate(int);
int alert_openlog(void);

int alert_init_ok=0;

void alert_cleanup(int code, void *priv)
{
	mesg(M_INFO,"alert: flushing logfiles");


	if ( !first_packet ) {
		/* If we have logged to this file then rotate it */
		alert_rotate(0);
	}else if ( alert_init_ok ) {
		/* If not, then unlink it */
		if ( unlink(fn_alert) )
			mesg(M_CRIT,"alert: unlink(): %s", get_err());
	}
}

void alert_open_fifo(void)
{
	int flags=0;

	if ( stormwall_mode==STORMWALL_FAIL ) {
		flags=O_NONBLOCK;
	}else if ( stormwall_mode==STORMWALL_WAIT ) {
		mesg(M_DEBUG,"alert: attempting to contact stormwall...");
	}else return;

	if ( (fifo_fd=open(fifo_fn, O_WRONLY|flags))<0 ) {
		mesg(M_ERR, "%s: open(): %s", fifo_fn, get_err());
		cleanup(EXIT_ERR, "Unable to contact stormwall");
		return;
	}

	if ( stormwall_mode==STORMWALL_WAIT ) {
		mesg(M_DEBUG,"alert: ...SUCCESS!");
	}
}

/* By sending a SIGHUP firestorm can cause
 * a log rotation at any time */
void alert_hup(void)
{
	if ( !first_packet ) {
		alert_rotate(1);
		last_bytes=sizeof(struct efile_hdr);
		last_time=0;
		first_packet=1;
	}
}

void alert_init(void) {cleanup_add(alert_cleanup, NULL);}

int cb_stormwall(struct arg *a, void *priv)
{
	if ( !strcmp("none", a->val.v_str) ) {
		stormwall_mode=STORMWALL_NONE;
	}else if ( !strcmp("wait", a->val.v_str) ) {
		stormwall_mode=STORMWALL_WAIT;
	}else if ( !strcmp("fail", a->val.v_str) ) {
		stormwall_mode=STORMWALL_FAIL;
	}else{
		mesg(M_ERR, "stormwall can be one of 'none', 'wait' or 'fail'");
		return 0;
	}

	return 1;
}

/* Config file hook for setting up output mechanisms */
void alert_conf_hook(char *args)
{
	struct arg the_args[]={
		{"stormwall", ARGTYPE_STRING, cb_stormwall},
		{"size", ARGTYPE_PBYTES, NULL, {vp_bytes:&max_bytes}},
		{"minutes", ARGTYPE_PUINT, NULL, {vp_uint:&max_time}},
		{NULL,ARGTYPE_NOP,NULL}
	};

	switch ( args && args_parse(the_args, args, NULL) ) {
	case -1:
		mesg(M_ERR,"logdir: parse error: %s", args);
	case 0: /* fall through */
		cleanup(EXIT_ERR,"logdir: failed");
		return;
	default:
		max_time*=60;
		break;
	}
}

void alert_conf_go(void)
{
	if ( !alert_openlog() ) {
		cleanup(EXIT_ERR, "there was a problem opening the logfile");
		return;
	}

	if ( max_bytes ) {
		mesg(M_INFO,"alert: rotating logs if files reach %uKB",
			max_bytes/1024);
	}

	if ( max_time ) {
		mesg(M_INFO,"alert: upper bound on log rotation is "
			"%u hrs %u mins", max_time/(60*60),
			(max_time%(60*60))/60);
	}

	if ( stormwall_mode==STORMWALL_NONE ) {
		mesg(M_INFO,"alert: stormwall support disabled - "
			"that means no remote alerting");
	}else{
		alert_open_fifo();
	}

	alert_init_ok=1;
}

/* Rotate a stale logfile */
int alert_rotate_old(char *fn)
{
	struct {
		struct efile_hdr fh;
		struct elog_pkthdr ph;
	}buf;
	int fd;

	if ( (fd=open(fn,O_RDONLY))<0 ) return 1;

	/* Read the date of the first packet and then rotate it */
	if ( read(fd, &buf, sizeof(buf)) <
		( ((char *)&buf.ph.h.ts.tv_usec - (char *)&buf)) +
		sizeof(buf.ph.h.ts.tv_usec) ) {
		close(fd);
		goto kill_it;
	}
	close(fd);
	if ( ntohs(buf.fh.magic)!=EF_MAGIC ) goto kill_it;

	rotate_time.tv_sec=buf.ph.h.ts.tv_sec;
	rotate_time.tv_usec=buf.ph.h.ts.tv_usec;

	/* Rotate it */
	alert_rotate(0);
	return 1;

kill_it:
	/* If there is no first packet, just unlink it */
	mesg(M_WARN, "alert: obliterating %s", fn);
	if ( unlink(fn) ) return 0;
	return 1;
}

/* Open a new logfile and write the header */
int alert_openlog(void)
{
	struct efile_hdr fhdr;
	int ret;
	int retry=0;

	/* Open up the log file */
try_again:
	if ( (log_fd=open(fn_alert,O_WRONLY|O_TRUNC|O_CREAT|O_EXCL,0600))<0 ) {
		if ( !retry && errno==EEXIST ) {
			if ( !alert_rotate_old(fn_alert) ) return 0;
			retry++;
			goto try_again;
		}
		mesg(M_CRIT,"%s: open(): %s", fn_alert, get_err());
		return 0;
	}

	/* Construct header */
	fhdr.magic=htonl(EF_MAGIC);
	fhdr.flags=0;
	fhdr.vers_minor=EF_VERS_MIN;
	fhdr.vers_major=EF_VERS_MAJ;

	/* Write header */
	ret=write(log_fd, &fhdr, sizeof(fhdr));
	if ( ret<sizeof(fhdr) ) {
		if ( ret<0 ) {
			mesg(M_CRIT,"elog: %s: write(): %s",
				fn_alert, get_err());
			return 0;
		}else{
			mesg(M_CRIT,"elog: %s: writing file header: "
				"truncated", fn_alert);
			return 0;
		}
	}
	return 1;
}

/* All-in-one function to perform a log rotation */
void alert_rotate(int new)
{
	struct stat st;
	char rotate_fn[32];

	/* Close old logfile */
	if ( fstat(log_fd, &st) ) mesg(M_CRIT,"alert: fstat(): %s", get_err());
	if ( fsync(log_fd) ) mesg(M_CRIT,"alert: fsync(): %s", get_err());
	if ( close(log_fd) ) mesg(M_CRIT,"alert: close(): %s", get_err());

	/* Build a name to rotate the old file to */
	sprintf(rotate_fn, "./log/@%.8lx.%.8lx.elog",
		rotate_time.tv_sec, st.st_ino);

	/* Rotate it */
	if ( rename(fn_alert, rotate_fn) )
		mesg(M_CRIT,"alert: rename(): %s", get_err());

	mesg(M_DEBUG,"logrotate: %s", rotate_fn);

	/* Wake up the stormwall daemon if necessary */
	if ( stormwall_mode ) {
		u_int8_t code=STORMWALL_ROTATE;
		if ( write(fifo_fd, (char *)&code, sizeof(code))<0 ) {
			mesg(M_CRIT, "alert: stormwall: %s", get_err());
		}
	}

	/* Open a new one */
	if ( new ) alert_openlog();

}

/* Token bucket rate-limiting filter */
int alert_ratelimit(struct packet *pkt, struct alert *a)
{
	unsigned long h;

	if ( !a->cost ) return 1;

	/* Build a single time value with 0.01s resolution */
	h=pkt->time.tv_sec * RATE_SEC;
	h+=pkt->time.tv_usec / (1000000UL/RATE_SEC);

	a->toks += h - a->last_msg;
	a->last_msg=h;

	if ( a->toks > a->burst )
		a->toks=a->burst;

	if ( a->toks >= a->cost ) {
		if ( a->missed )
			mesg(M_DEBUG, "alert: %s: missed %i alerts", a->alert, a->missed);
		a->missed=0;
		a->toks -= a->cost;
		return 1;
	}

	a->missed++;
	return 0;
}

/* This is where alerts come to */
void alert(struct generator *gen, struct packet *pkt, struct alert *a)
{
	static char decode_buf[4096];
	char *ptr_end=decode_buf+sizeof(decode_buf);
	struct elog_dhdr *dh;
	char *ptr=decode_buf;
	struct elog_pkthdr ph;
	struct iovec iov[6];
	char padding[4]={0,0,0,0};
	int i=0;
	int j;

	if ( !pkt ) return;

	if ( !alert_ratelimit(pkt,a) )
		return;

	/* Construct record header */
	ph.h.reclen=sizeof(ph)+pkt->caplen;
	ph.h.type=ELOG_ALERT;
	ph.h.prio=a->priority;
	ph.h.reserved=0;
	ph.h.ts.tv_sec=pkt->time.tv_sec;
	ph.h.ts.tv_usec=pkt->time.tv_usec;

	/* Construct alert specific bit */
	ph.sid=a->sid;
	ph.rev=a->rev;
	ph.pflags=pkt->flags;
	ph.decode_len=0;
	ph.gen_len=strlen(gen->name)+1; ph.h.reclen+=ph.gen_len;
	ph.alert_len=strlen(a->alert)+1; ph.h.reclen+=ph.alert_len;
	ph.pkt_len=pkt->len;
	ph.pkt_caplen=pkt->caplen;

	/* Record header vector */
	iov[i].iov_base=&ph;
	iov[i++].iov_len=sizeof(ph);

	/* Construct the decode data */
	/* TODO: Call plugins to serialise state data */
	for(j=0; j<pkt->llen; j++) {
		int nlen, real;
	
		if ( pkt->layer[j].proto ) {
			real=nlen=strlen(pkt->layer[j].proto->name)+1;
		}else{
			real=nlen=0;
		}

		/* Round to the nearest multiple of 4 */
		if ( nlen & 0x3 ) nlen+=(4-(nlen&0x3));

		if ( ptr+nlen+sizeof(*dh) > ptr_end ) {
			mesg(M_CRIT,"elog: decode buffer too small!");
			return;
		}

		/* Fill in the header */
		dh=(struct elog_dhdr *)ptr;
		ptr+=sizeof(*dh);
		dh->tot_len=(sizeof(*dh)+nlen)>>2;
		dh->name_len=nlen>>2;
		dh->pkt_ofs=htons(pkt->layer[j].h.raw - pkt->base);
		dh->flags=htonl(pkt->layer[j].flags);

		/* Then the protocol name  */
		if ( real ) {
			memcpy(ptr, pkt->layer[j].proto->name, real);
			ptr+=real;
			if ( nlen-real ) {
				memset(ptr, 0, nlen-real);
				ptr+=nlen-real;
			}
		}
	}

	/* decode data vector */
	iov[i].iov_base=decode_buf;
	iov[i].iov_len=ptr-decode_buf;

	/* fill in/update record header fields */
	ph.decode_len=(iov[i].iov_len>>2);
	ph.h.reclen+=iov[i++].iov_len;

	/* raw packet data vector */
	iov[i].iov_base=pkt->base;
	iov[i++].iov_len=pkt->caplen;

	/* generator and alert string vectors */
	iov[i].iov_base=gen->name;
	iov[i++].iov_len=ph.gen_len;
	iov[i].iov_base=a->alert;
	iov[i++].iov_len=ph.alert_len;

	/* Pad to multiples of 4 bytes (ensures
	 * aligned access to next header if file
	 * is mmap()ed in) */
	if ( ph.h.reclen & 3 ) {
		iov[i].iov_base=padding;
		iov[i].iov_len=(4-(ph.h.reclen&3));
		ph.h.reclen+=iov[i++].iov_len;
	}

	/* Check to see if we exceeded max logfile size or
	 * haven't rotated in a while */
	if ( !last_time ) last_time=pkt->time.tv_sec;
	if ( (max_bytes && (last_bytes+ph.h.reclen > max_bytes)) ||
		(max_time && (pkt->time.tv_sec > last_time+max_time)) ) {
		alert_rotate(1);
		last_bytes=sizeof(struct efile_hdr);
		last_time=pkt->time.tv_sec;
		first_packet=1;
	}
	last_bytes+=ph.h.reclen;

	/* Convert to network byte order  */
	ph.h.reclen=htonl(ph.h.reclen);
	ph.h.type=htons(ph.h.type);
	ph.h.ts.tv_sec=htonl(ph.h.ts.tv_sec);
	ph.h.ts.tv_usec=htonl(ph.h.ts.tv_usec);
	ph.sid=htonl(ph.sid);
	ph.rev=htonl(ph.rev);
	ph.pflags=htonl(ph.pflags);
	ph.decode_len=htons(ph.decode_len);
	ph.pkt_len=htonl(ph.pkt_len);
	ph.pkt_caplen=htonl(ph.pkt_caplen);

	/* Rotation filenames contain the timestamp
	 * and serial number of the first packet in
	 * that file */
	if ( first_packet ) {
		rotate_time.tv_sec=pkt->time.tv_sec;
		rotate_time.tv_usec=pkt->time.tv_usec;
		first_packet=0;
	}

	/* Actually write it all */
	if ( writev(log_fd, iov, i)<0 ) {
		mesg(M_CRIT,"elog: %s: writev(): %s", fn_alert, get_err());
	}
}
