/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2002 Gianni Tedesco
 * This program is released under the terms of the GNU GPL version 2
 *
 *	This file handles firestorm-nids configuration from both the command
 *	line and from the configuration file.
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <getopt.h>

#include <firestorm.h>
#include <cleanup.h>
#include <template.h>
#include <loader.h>
#include <capture.h>
#include <packet.h>
#include <generator.h>
#include <alert.h>
#include <signature.h>
#include <parser.h>
#include <preproc.h>
#include <decode.h>
#include <conf.h>

#ifndef _BSD_SOURCE
#define _BSD_SOURCE
#endif
#include <grp.h>

#define CONF_MAXLEN 1024

#define terminator(x) ( (x)=='\n' || (x)=='\r' )
#define whitespace(x) ( (x)==' ' || (x)=='\t' )

/* Linked list of config lines */
#define LT_LOG 0
#define LT_PREPROC 1
#define LT_SIG 2
struct confline {
	struct confline *next;
	int ln;
	char *args;
	char *args2;
};
struct confline *conf_head[3]={NULL,NULL,NULL};
struct confline *conf_tail[3]={NULL,NULL,NULL};

/* Profile naming */
static char default_profile[]="default";
char *profile_name=NULL;
char *fn_config=NULL;
char *fn_alert=NULL;
char *fn_log=NULL;

/* Command line parameters */
struct option opts[]={
	{"profile", 1, 0, 'p'},
	{"uid", 1, 0, 'u'},
	{"gid", 1, 0, 'g'},
	{"root", 1, 0, 'r'},
	{"capdev", 1, 0, 'c'},
	{"options", 1, 0, 'o'},
	{"plugins", 1, 0, 'L'},
	{"plugin", 1, 0, 'l'},
	{NULL, 0, 0, 0}
};

/* Perform initial file descriptor tweaks */
void conf_tweak_fds()
{
	int fds=sysconf(_SC_OPEN_MAX);
	while(fds>2) close(fds--);

	if ( (fds=open("/dev/null", O_RDWR))<0 ) {
		cleanup(EXIT_ERR, "/dev/null: open(): %s", get_err());
	}

	/* Make stdin and stdout be /dev/null */
	if ( fds!=0 && dup2(fds,0)<0 )
		cleanup(EXIT_ERR, "stdin: dup2(): %s", get_err());
	if ( fds!=1 && dup2(fds,1)<0 )
		cleanup(EXIT_ERR, "stdout: dup2(): %s", get_err());

	/* if fds<2 it means stdin/stdout/stderr weren't
	 * open to start with so dup stderr aswell */
	if ( fds<2 && dup2(fds,2)<0 )
		cleanup(EXIT_ERR, "stderr: dup2(): %s", get_err());

	if ( fds>2 ) close(fds);
}

/* Use the profile name to create our file paths */
void conf_setup_names(void)
{
	int nlen;
	static char fnp_root[]="profile.";
	static char fnp_config[]="firestorm.conf";
	static char fnp_alert[]="alert.elog";
	static char fnp_log[]="messages";

	nlen=strlen(profile_name)+sizeof(fnp_root);

	if ( !(fn_config=malloc(nlen+sizeof(fnp_config))) ) {
		cperror("malloc");
	}
	if ( !(fn_alert=malloc(nlen+sizeof(fnp_alert))) ) {
		cperror("malloc");
	}
	if ( !(fn_log=malloc(nlen+sizeof(fnp_log))) ) {
		cperror("malloc");
	}

	sprintf(fn_config, "%s%s/%s", fnp_root, profile_name, fnp_config);
	sprintf(fn_alert, "%s%s/%s", fnp_root, profile_name, fnp_alert);
	sprintf(fn_log, "%s%s/%s", fnp_root, profile_name, fnp_log);
}

void conf_background(void)
{
	switch(fork()) {
		case 0: break;
		case -1:
			cleanup(EXIT_ERR, "fork(): %s", get_err());
		default: _exit(0);
	}

	if (setsid()<0) {
		cleanup(EXIT_ERR, "setsid(): %s", get_err());
	}

	switch(fork()) {
		case 0: break;
		case -1:
			cleanup(EXIT_ERR, "fork(): %s", get_err());
		default: _exit(0);
	}

	if ( dup2(0,2)<0 )
		cleanup(EXIT_ERR, "stderr: dup2(): %s", get_err());
}

void conf_cmdline(int argc, char **argv)
{
	int c,opti;
	uid_t uid=0;
	uid_t gid=0;
	char *root_dir=NULL;
	char *capdev=NULL;
	char *capdev_ops=NULL;

	conf_tweak_fds();

	while( (c=getopt_long(argc,argv,"p:u:g:r:c:o:L:l:",
		(const struct option *)&opts, &opti))!=-1 ) {
		switch(c) {
		case 'p':
			profile_name=optarg;
			break;
		case 'u':
			if ( strtouint(optarg,&uid) ) goto err;
			break;
		case 'g':
			if ( strtouint(optarg,&gid) ) goto err;
			break;
		case 'r':
			root_dir=optarg;
			break;
		case 'c':
			capdev=optarg;
			break;
		case 'o':
			capdev_ops=optarg;
			break;
		case 'L':
			loader_load_dir(optarg);
			break;
		case 'l':
			if ( !loader_load(optarg) ) goto err;
			break;
		default:
			goto err;
			break;
		}
	}

	/* Check arguments */
	if ( !root_dir || !capdev ) goto err;
	if ( !profile_name ) profile_name=default_profile;

	/* All plugins should be loaded so lets go */
	decode_resolve_all();

	/* Change to root dir */
	if ( chdir(root_dir) ) {
		cleanup(EXIT_ERR, "%s: chdir(): %s", root_dir, get_err());
	}

	/* Initialise captures */
	capture_add(capdev, capdev_ops);

	if ( geteuid() ) {
		mesg(M_WARN, "cannot chroot, im not root");
	}else{
		/* Chroot in to our root dir */
		if ( chroot(".") ) {
			cleanup(EXIT_ERR, "%s: chroot(): %s",
				root_dir, get_err());
		}

		/* You should never need to do this */
		if ( uid==0 || gid==0 ) {
			cleanup(EXIT_ERR, "I will NOT run as root you moron!");
		}

		/* Drop the GID */
		if ( setgroups(1,&gid)==-1 || setgid(gid)==-1 ) {
			cleanup(EXIT_ERR, "unable to drop gid to %u: %s",
				gid, get_err());
		}
	
		/* Drop the UID */
		if ( setuid(uid) == -1 ) {
			cleanup(EXIT_ERR, "unable to drop uid to %u: %s",
				uid, get_err());
		}
	}

	conf_setup_names();

	return;
err:
	cleanup(EXIT_ERR,"Bad arguments: %s should not "
		"be run by hand", argv[0]);
}

void conf_line(char *fn, int ln, char *line)
{
	char *cur;
	char *args=NULL;
	char *args2=NULL;
	int state=0;
	int ok=0;
	int type=-1;
	struct confline *x;

	for(cur=line; *cur; cur++) {
		if ( terminator(*cur) ) {
			*cur=' ';
			ok=1;
		}

		if ( state==0 && whitespace(*cur) ) {
			*cur=0;
			state=1;
			if (!strcmp(line, "log")) {
				type=LT_LOG;
			}else if (!strcmp(line, "preprocessor")){
				type=LT_PREPROC;
			}else if (!strcmp(line, "signatures")){
				type=LT_SIG;
			}else{
				cleanup(EXIT_ERR,"%s:%i: %s: unknown keyword",
					fn,ln,line);
			}
		}else if ( state==1 && !whitespace(*cur) ) {
			args=cur;
			if ( type==LT_LOG ) {
				state=99;
			}else{
				state=2;
			}
		}else if ( state==2 && whitespace(*cur) ) {
			*cur=0;
			state=3;
		}else if ( state==3 && !whitespace(*cur) ) {
			args2=cur;
			state=4;
		}

		if ( ok ) {
			*cur=0;
			break;
		}
	}

	/* Check for errors */
	if (!ok) cleanup(EXIT_ERR, "%s:%i: line too long",fn,ln);
	if (!args) goto parse_err;
	if (type<0) goto parse_err;

	if ( !(x=calloc(1, sizeof(*x))) )
		cperror("calloc");

	x->ln=ln;

	if ( args && !(x->args=strdup(args)) )
		cperror("calloc");

	if ( args2 && !(x->args2=strdup(args2)) )
		cperror("calloc");

	if ( !conf_head[type] ) {
		conf_head[type]=x;
		conf_tail[type]=x;
	}else{
		if ( !type ) {
			cleanup(EXIT_ERR, "%s:%i: 'log' directive "
				"specified more than once", fn, ln);
		}
		conf_tail[type]->next=x;
		conf_tail[type]=x;
	}

	return;
parse_err:
	cleanup(EXIT_ERR, "%s:%i: parse error",fn,ln);
}

void conf_local_config(void)
{
	FILE *f;
	int line=0;
	char buf[CONF_MAXLEN];
	struct confline *c;

	/* Open the config file */
	if ( (f=fopen(fn_config,"r"))==NULL ) {
		cleanup(EXIT_ERR, "%s: fopen(): %s", fn_config, get_err());
	}

	/* Parse the thing */
	while(fgets(buf,sizeof(buf),f)) {
		char *p=buf;
		line++;

		while(*p && whitespace(*p) ) p++;
		if ( terminator(*p) || *p=='#' ) continue;
		conf_line(fn_config, line, p);
	}

	/* Close it */
	fclose(f);

	/* Now lets actually do the business */
	c=conf_head[LT_LOG];
	alert_conf_hook(c->args);
	free(c->args);
	free(conf_head[LT_LOG]);

	/* Load all the preprocessors */
	for(c=conf_head[LT_PREPROC]; c; ) {
		struct confline *tmp=c;
		c=c->next;
		if ( !preproc_load(tmp->args, tmp->args2) ) {
			cleanup(EXIT_ERR, "%s:%i: preprocessor failed",
				fn_config, tmp->ln);
		}
		if (tmp->args) free(tmp->args);
		if (tmp->args2) free(tmp->args2);
		free(tmp);
	}

	/* Load all the signatures */
	for(c=conf_head[LT_SIG]; c; ) {
		struct confline *tmp=c;
		c=c->next;
		if ( !parser_load(tmp->args, tmp->args2) ) {
			cleanup(EXIT_ERR, "%s:%i: signatures failed",
				fn_config, tmp->ln);
		}
		if (tmp->args) free(tmp->args);
		if (tmp->args2) free(tmp->args2);
		free(tmp);
	}
}
