/*
* This file is part of Firestorm NIDS
* Copyright (c) 2002 Gianni Tedesco
* This program is released under the terms of the GNU GPL version 2
*
*    This is the decode engine. We keep a big internal graph of all
*    protocols and all their dependancies on each other. The actual
*    decoding is done by plugins.
*/
#include <firestorm.h>
#include <rlim.h>
#include <string.h>
#include <packet.h>
#include <cleanup.h>
#include <preproc.h>
#include <generator.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <string_hash.h>

/* libpcap DLT_* protocol */
struct proto pcap_dlt_p=init_proto("__pcap_dlt", NULL, NULL);

/* Global protocol list */
#define DHASH_SIZE 128
struct proto *protos[DHASH_SIZE];

//struct proto *protos=&pcap_dlt_p;
unsigned int num_protos=0;
unsigned int proto_warn=0;

/* Protocol request list */
struct preq_list *preqs=NULL;

void decode_cleanup(int code, void *priv)
{
	struct preq_list *foo, *bar;

	for(foo=preqs; foo;) {
		bar=foo;
		foo=foo->next;
		free(bar);
	}

}

void decode_init(void)
{
	u_int32_t h;

	memset(protos, 0, sizeof(protos));
	h=string_hash((char *)pcap_dlt_p.name)&(DHASH_SIZE-1);
	protos[h]=&pcap_dlt_p;

	cleanup_add(decode_cleanup, NULL);
}

/* Find a protocol by name */
struct proto *decode_proto(char *proto)
{
	struct proto *p;
	for(p=protos[string_hash(proto)&(DHASH_SIZE-1)]; p; p=p->next) {
		if ( !strcmp(p->name, proto) ) { 
			return p;
		}
	}
	return NULL;
}

/* Capture plugins call this   */
struct proto *decode_subproto(char *proto, unsigned int lt)
{
	struct proto *p;
	struct proto_child *c;

	if ( !(p=decode_proto(proto)) ) return NULL;

	for(c=p->children; c; c=c->next) {
		if ( c->id == lt ) {
			return c->proto;
		}
	}
	return NULL;
}

/* plugins call this to register protocols */
int decode_add(struct proto *p, struct proto_req *r)
{
	struct preq_list *req;
	u_int32_t h;

	if ( !p || !r || !p->name || !p->decode ) return 0;

	if ( strlen(p->name) > PROTO_LEN ) {
		mesg(M_WARN, "decode: %s: protocol name is too long", p->name);
	}

	if ( num_protos >= MAX_PROTOS ) {
		if ( !proto_warn ) {
			proto_warn++;
			mesg(M_WARN,"decode: MAX_PROTOS exceeded.");
		}
		return 0;
	}

	/* Check if we have already been added */
	if ( p->next || p->children || decode_proto((char *)p->name) ) {
		mesg(M_DEBUG,"decode: error: '%s' protocol is "
			"already registered", p->name);
		return 0;
	}

	if ( !(req=calloc(1, sizeof(struct preq_list))) ) {
		cperror("calloc");
	}

	/* Save the requests for later, when all the protocols
	 * are done registering. We resolve them in
	 * decode.c::decode_resolve_all() */
	req->proto=p;
	req->arr=r;
	req->next=preqs;
	preqs=req;

	h=string_hash((char *)p->name)&(DHASH_SIZE-1);

	/* Add to the list */
	p->next=protos[h];
	protos[h]=p;
	num_protos++;

	return 1;
}

/* Resolve a protocol request */
int decode_resolve(struct proto_req *req, struct proto *proto)
{
	struct proto_child *pc;
	struct proto *parent;

	if ( !(parent=decode_proto(req->name)) ) {
		return 0;
	}

	if ( !(pc=calloc(1, sizeof(struct proto_child))) ) {
		cperror("calloc");
	}

	/* fill in the proto_child */
	pc->id=req->id;
	pc->proto=proto;

	/* append it to the rightful owner */
	pc->next=parent->children;
	parent->children=pc;

#if 0
	mesg(M_DEBUG,"resolve: %s:%u -> %s",
		parent->name,
		pc->id,
		pc->proto->name);
#endif

	return 1;
}

/* Build the graph of inter-dependencies
 * for all registered protocols */
int decode_resolve_all(void)
{
	struct preq_list *foo, *bar;
	int i;

	for(foo=preqs; foo;) {
		bar=foo;
		foo=foo->next;

		for(i=0; bar->arr[i].name; i++) {
			decode_resolve(&bar->arr[i], bar->proto);
		}

		/* this is only a temporary structure
		 * after all */
		free(bar);
	}

	preqs=NULL;

	return 1;
}
