/*
* This file is part of Firestorm NIDS
* Copyright (c) 2002 Gianni Tedesco
* This program is released under the terms of the GNU GPL version 2
*
* This file provides a simple API for reading elog files. It only
* works for files <2GB right now and isn't very secure. Don't read
* untrusted data files just yet.
*/
#include <stdlib.h>
#include <stdio.h>
#include <netinet/in.h>

#include <firestorm.h>
#include <packet.h>
#include <generator.h>
#include <alert.h>
#include <signature.h>
#include <target.h>
#include <capture.h>
#include <capdev.h>
#include <cleanup.h>
#include <decode.h>
#include <elog.h>

struct eread_priv {
	proc_target alert_fn;
	void *alert_priv;
	char *name;
};

struct capdev null_capdev={NULL,"elog_read",NULL,NULL,NULL};
struct cap_s null_cap={&null_capdev,CAP_STATE_CAPTURE,NULL,NULL};

/* Create a new elog parsing structure */
void *elog_new(char *fn)
{
	struct eread_priv *p;
	if ( !(p=calloc(1,sizeof(struct eread_priv))) ) {
		mesg(M_ERR, "elog_new(): %s", get_err());
	}
	p->name=fn;
	return p;
}

/* Free the created structure */
void elog_free(void *p)
{
	if ( !p ) return;
	free(p);
}

/* Hook up a function to recieve alerts */
void elog_set_alert_notifier(void *rp, proc_target alert_fn, void *t_priv)
{
	struct eread_priv *e=rp;
	if ( !alert_fn ) return;
	e->alert_fn=alert_fn;
	e->alert_priv=t_priv;
}

/* Parse alert decode data */
int elog_decode(struct packet *pkt, char *buf, int buflen)
{
	struct elog_dhdr *d;
	char *end=buf+buflen;
	struct proto *p;

	while(buf+sizeof(*d)<=end) {
		d=(struct elog_dhdr *)buf;

		if ( (d->tot_len<<2)<sizeof(*d) ) return 0;
		if ( buf+(d->tot_len<<2) > end )
			return 0;

		if ( d->name_len==0 ) {
			p=NULL;
		}else{
			p=decode_proto(buf+sizeof(*d));
		}

		pkt->layer[pkt->llen].h.raw=
			pkt->base+ntohs(d->pkt_ofs);
		pkt->layer[pkt->llen].proto=p;
		pkt->layer[pkt->llen].flags=ntohl(d->flags);
		pkt->layer[pkt->llen++].session=NULL;
	
		buf+=d->tot_len<<2;
	}
	return 1;
}

/* Handler for alert records */
int elog_handle_alert(struct eread_priv *e,
	struct elog_pkthdr *ph,
	char *decode_buf, char *pkt_data,
	char *str_g, char *str_a)
{
	struct generator gen;
	struct packet pkt;
	struct alert a;

	/* Fill in generator and alert structures */
	gen.next=NULL;
	gen.name=str_g;
	a.alert=str_a;
	a.sid=ntohl(ph->sid);
	a.rev=ntohl(ph->rev);
	a.priority=ph->h.prio;

	/* Fill in packet structure */
	pkt.serial=0;
	pkt.len=ntohl(ph->pkt_len);
	pkt.caplen=ntohl(ph->pkt_caplen);
	pkt.llen=0;
	pkt.base=pkt_data;
	pkt.end=pkt_data+pkt.caplen;
	pkt.time.tv_sec=ntohl(ph->h.ts.tv_sec);
	pkt.time.tv_usec=ntohl(ph->h.ts.tv_usec);
	pkt.capture=&null_cap;
	pkt.flags=ntohl(ph->pflags);

	/* Fill in decode data */
	if ( !elog_decode(&pkt, decode_buf,
		ntohs(ph->decode_len)<<2) ) {
		return 0;
	}

	/* Call the output plugin */
	null_cap.args=e->name;
	return e->alert_fn ? e->alert_fn(&gen, &pkt, &a, e->alert_priv) : 1;
}

/* Parse a memory mapped elog file */
int elog_run_mmap(void *p, char *begin, char *end)
{
	char *cur=begin;
	struct efile_hdr *fh;
	struct elog_pkthdr *ph;
	struct eread_priv *e=p;

	if ( !p ) return 0;

	/* Check the header */
	fh=(struct efile_hdr *)cur;
	cur+=sizeof(*fh);
	if ( cur > end ) return 0;
	if ( ntohl(fh->magic)!=EF_MAGIC ) {
		mesg(M_ERR,"%s: Bad voodoo magic", e->name);
		return 0;
	}

	/* Loop through each packet */
	while(cur+(sizeof(*ph))<end) {
		char *decode_buf;
		char *pkt_data;
		char *str_g;
		char *str_a;

		/* We are guaranteed the header */
		ph=(struct elog_pkthdr *)cur;
		decode_buf=cur+sizeof(*ph);
		pkt_data=decode_buf+(ntohs(ph->decode_len)<<2);
		str_g=pkt_data+(ntohl(ph->pkt_caplen));
		str_a=str_g+ph->gen_len;

		/* Quick integrity check */
		if ( ntohl(ph->h.reclen) < sizeof(*ph) ) return 0;
		if ( cur+ntohl(ph->h.reclen) > end ) return 0;
		if ( str_a+ph->alert_len > end ) return 0;

		/* Dispatch this record to
		 * the appropriate handler */
		switch( ntohs(ph->h.type) ) {
		case ELOG_ALERT:
			if ( !elog_handle_alert(e,
				ph, decode_buf,
				pkt_data, str_g, str_a) ) {
				return 0;
			}
			break;
		default:
			break;
		}
		cur+=ntohl(ph->h.reclen);
	}

	return 1;
}
