/*
* This file is part of Firestorm NIDS
* Copyright (c) 2002 Gianni Tedesco
* This program is released under the terms of the GNU GPL version 2
*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <getopt.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <netinet/in.h>

#include <firestorm.h>
#include <packet.h>
#include <cleanup.h>
#include <loader.h>
#include <generator.h>
#include <alert.h>
#include <args.h>
#include <target.h>
#include <signature.h>
#include <decode.h>
#define __PLUGIN__
#include <matcher.h>
#undef __PLUGIN__
#include <detect.h>
#include <capture.h>
#include <capdev.h>
#include <elog.h>

struct target *t=NULL;
void *t_priv=NULL;
char *fn=NULL;
char *cmd=NULL;

/* Functions we may need to fake */
int preproc_activate(char *n, void *p1, void *p2, void *p3){return 1;}
int preproc_add(char *n, void *p1, void *p2){return 1;}
void *matcher_find(char *n){return NULL;}
int generator_add(struct generator *g){return 1;}
void serial_number(serial_t *s){*s=0;}
void alert(struct generator *g, struct packet *p, struct alert *a){}

/* TODO: Get the full set on here */
struct {
	const char *name;
	void *data;
}exports[]={
	{"target.add", 		target_add},
	{"args.parse", 		args_parse},
	{"decode.add", 		decode_add},
	{"decode.subproto", 	decode_subproto},
	{"decode.proto", 	decode_proto},
	{"decode.__pcap_dlt", 	&pcap_dlt_p},
	{"mesg",		mesg},
	{"preproc.activate",	preproc_activate},
	{"preproc.add",		preproc_add},
	{"matcher.find",	matcher_find},
	{"generator.add",	generator_add},
	{"serial_number",	serial_number},
	{"alert",		alert},
	{"detect.add_sig",	detect_add_sig},
	{"detect.free_sig",	detect_free_sig},
	{"detect.set",		detect_set},
	{"detect",		detect},
	{NULL,NULL}
};


/* Only print wanings,errors,and critical messages */
void mesg(unsigned char code, char *fmt, ...)
{
	va_list va;
	if ( code > M_WARN ) return;
	va_start(va, fmt);
	/* Use stderr cos output may use stdout */
	fprintf(stderr, "%s: ", cmd);
	vfprintf(stderr, fmt, va);
	fprintf(stderr, "\n");
	va_end(va);
}

/* Process a file */
void do_file(int fd)
{
	struct stat st;
	char *map, *end;
	void *ep;

	if ( fstat(fd, &st) ) {
		mesg(M_ERR, "%s: fstat(): %s",
			fn, get_err());
		cleanup(EXIT_ERR,NULL);
	}

	if ( st.st_size<sizeof(struct efile_hdr) ) {
		mesg(M_ERR, "%s: file is too small!",fn);
		cleanup(EXIT_ERR,NULL);
	}

	if ( (map=mmap(NULL,st.st_size,
		PROT_READ,MAP_SHARED,fd,0))==MAP_FAILED ) {
		mesg(M_ERR, "%s: mmap(): %s", fn, get_err());
		cleanup(EXIT_ERR,NULL);
	}

	end=map+st.st_size;

	if ( !(ep=elog_new(fn)) ) return;
	elog_set_alert_notifier(ep, t->target, t_priv);
	if ( !elog_run_mmap(ep, map, end) ) {
		elog_free(ep);
		cleanup(EXIT_ERR, NULL);
	}
	elog_free(ep);
}

void *plugin_callback(const char *item)
{
	u_int32_t i;

	for(i=0; exports[i].name; i++) {
		if ( !strcmp(exports[i].name, item) ) {
			return exports[i].data;
		}
	}

	fprintf(stderr, "resource: request for unknown object: '%s'\n", item);
	return NULL;
}

struct option opts[]={
	{"plugin-dir", 1, 0, 'p'},
	{"in", 1, 0, 'i'},
	{"format", 1, 0, 'f'},
	{"options", 1, 0, 'o'},
	{"help", 0, 0, 'h'},
	{NULL, 0, 0, 0}
};

void usage(int ecode)
{
	fprintf(stderr, "\n  Usage: %s -i infile -f format [-o 'options']\n\n", cmd);
	fprintf(stderr, "    --in,-i          Specify the input file\n");
	fprintf(stderr, "    --format,-f      Choose an output format\n");
	fprintf(stderr, "    --options,-o     Plugin specific option string\n");
	fprintf(stderr, "    --plugin-dir,-p  Load plugins from this directory\n");
	fprintf(stderr, "    --help,-h        This message\n\n");
	cleanup(ecode,NULL);
}

int main(int argc, char **argv)
{
	int c,opti;
	char *infile=NULL;
	char *oformat=NULL;
	char *p_options=NULL;

	/* Initialise what we need to */
	cmd=argv[0];
	loader_init(plugin_callback);

	/* Initialise subsystems */
	decode_init();

	/* Do command line args */
	while( (c=getopt_long(argc, argv, "p:i:f:o:h?",
		(const struct option *)&opts, &opti))!=-1 ) {
		switch (c) {
			case 'p':
				loader_load_dir(optarg);
				break;
			case 'i':
				if ( infile ) usage(EXIT_ERR);
				infile=optarg;
				break;
			case 'f':
				if ( oformat ) usage(EXIT_ERR);
				oformat=optarg;
				break;
			case 'o':
				if ( p_options ) usage(EXIT_ERR);
				p_options=optarg;
				break;
			case 'h':
			case '?':
				usage(EXIT_OK);
				break;
			default:
				usage(EXIT_ERR);
		};
	}

	loader_load_dir(PLUGIN_DIR "/protocols");
	loader_load_dir(PLUGIN_DIR "/output");

	decode_resolve_all();

	if ( !oformat ) {
		mesg(M_ERR, "No output format specified");
		cleanup(EXIT_ERR,NULL);
	}

	/* Get the right output plugin */
	if ( !(t=target_find(oformat)) ) {
		mesg(M_ERR, "%s: Cannot find specified output format",
			oformat);
		cleanup(EXIT_ERR,NULL);
	}

	/* Initialise it with user specified arguments*/
	if ( !t->validate(p_options, &t_priv) ) {
		mesg(M_ERR, "%s: Invalid options for output plugin",
			oformat);
		cleanup(EXIT_ERR,NULL);
	}

	/* Process the input file */
	if ( infile ){
		int fd;

		fn=infile;

		/* Open the input file */
		if ( (fd=open(infile,O_RDONLY))<0 ) {
			mesg(M_ERR, "%s: open(): %s",
				infile, get_err());
			cleanup(EXIT_ERR,NULL);
		}

		do_file(fd);
	}else{
		/* Read from stdin */
		fn="stdin";
		do_file(0);
	}

	cleanup(EXIT_OK,NULL);
	return 0;
}
