/*
* This file is part of Firestorm NIDS
* Copyright (c) 2002 Gianni Tedesco
* This program is released under the terms of the GNU GPL version 2
*
*    This file is the entry point of the firestorm sensor;
*    it is responsible for initialising all other components.
*    It is also where the plugin import callback routine lives.
*/
#include <config.h>
#include <time.h>
#include <signal.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <fcntl.h>

#include <config.h>
#include <firestorm.h>
#include <packet.h>
#include <conf.h>
#include <generator.h>
#include <alert.h>
#include <capdev.h>
#include <capture.h>
#include <cleanup.h>
#include <loader.h>
#include <parser.h>
#include <preproc.h>
#include <template.h>
#include <signature.h>
#include <matcher.h>
#include <decode.h>
#include <sensor.h>
#include <args.h>
#include <detect.h>

int mesg_mode=-1;
char *pbuf, *pbuf_orig;
int pbuflen,pbuflen_orig;

/* Initialised forwards, cleaned up backwards */
proc_subsys_init subsys_init[]={
	capdev_init,
	parser_init,
	decode_init,
	preproc_init,
	capture_init,
	alert_init,
	NULL
};

/* Resources to export to plugins */
struct {
	const char *name;
	void *data;
}exports[]={
	{"capdev.find", 	capdev_find},
	{"capdev.add", 		capdev_add},
	{"decode.add", 		decode_add},
	{"decode.subproto", 	decode_subproto},
	{"decode.proto", 	decode_proto},
	{"decode.__pcap_dlt", 	&pcap_dlt_p},
	{"parser.add", 		parser_add},
	{"parser.load", 	parser_load},
	{"preproc.add", 	preproc_add},
	{"preproc.find",	preproc_find},
	{"preproc.activate", 	preproc_activate},
	{"signature.rule",	signature_rule},
	{"signature.abort", 	signature_abort},
	{"signature.criteria",	signature_criteria},
	{"signature.modifier",	signature_modifier},
	{"signature.alert", 	signature_alert},
	{"signature.commit",	signature_commit},
	{"matcher.add",		matcher_add},
	{"matcher.find",	matcher_find},
	{"template.ptr",	template_ptr},
	{"template.ptrptr",	template_ptrptr},
	{"template.longrange",	template_longrange},
	{"template.shortrange",	template_shortrange},
	{"template.cmp_lr",	template_cmp_lr},
	{"template.cmp_sr",	template_cmp_sr},
	{"alert",		alert},
	{"generator.add",	generator_add},
	{"serial_number",	serial_number},
	{"args.parse",		args_parse},
	{"mesg",		mesg},
	{"detect.add_sig",	detect_add_sig},
	{"detect.free_sig",	detect_free_sig},
	{"detect.set",		detect_set},
	{"detect",		detect},
	{NULL, NULL}
};

void *plugin_callback(const char *item)
{
	u_int32_t i;

	for(i=0; exports[i].name; i++) {
		if ( !strcmp(exports[i].name, item) ) {
			return exports[i].data;
		}
	}

	mesg(M_ERR,"resource: request for unknown object: '%s'", item);
	return NULL;
}

/* Prepare for the log swapover */
extern char *fn_log;
FILE *mesg_file=NULL;
void mesg_swap(void)
{
	int fd;

	if ( (fd=open(fn_log, O_WRONLY|O_CREAT|O_TRUNC, 00640))<0 ) {
		mesg(M_ERR, "%s: open(): %s", fn_log, get_err());
		goto err;
	}

	if ( !(mesg_file=fdopen(fd, "w")) ) {
		mesg(M_ERR, "%s: fdopen(): %s", fn_log, get_err());
		goto err;
	}

	fwrite(pbuf_orig, pbuflen_orig-pbuflen, 1, mesg_file);
	free(pbuf_orig);
	fflush(mesg_file);
	return;
err:
	cleanup(EXIT_ERR, "Unable to open logfile for writing");
}

/* Print out a firestorm internal system log message */
void mesg(unsigned char code, char *fmt, ...)
{
	static char buf[1024];
	struct timeval tv;
	va_list va;
	char *cstr[]={
		"undefined",
		"critical",
		"error",
		"warning",
		"info",
		"debug"
	};

	gettimeofday(&tv, NULL);

	if ( code > M_MAX ) code=0;

	va_start(va, fmt);
	vsnprintf(buf, sizeof(buf), fmt, va);
	va_end(va);

	if ( mesg_mode<0 ) {
		int ret;

		if ( code < M_INFO ) {
			fprintf(stderr, "%s: %s\n", cstr[code], buf);
			fflush(stderr);
		}

		/* Try write to buffer */
		ret=snprintf(pbuf, pbuflen, "%.8lu.%.6lu %s: %s\n",
			tv.tv_sec, tv.tv_usec, cstr[code], buf);

		/* Uh-oh ran out of space */
		/* TODO: Realloc */
		if ( ret<0 ) {
			fprintf(stderr, "error: some logs msgs be missing\n");
			return;
		}

		pbuflen-=ret;
		pbuf+=ret;

		/* snprintf will happily chew on negative numbers */
		if ( pbuflen<0 ) pbuflen=0;
	}else{
		fprintf(mesg_file, "%.8lu.%.6lu %s: %s\n",
			tv.tv_sec, tv.tv_usec, cstr[code], buf);
		fflush(mesg_file);
	}
}

#ifdef HAVE_SIGACTION 
void deadly_signal(int signum, siginfo_t *info, void *foo)
{
	mesg(M_INFO,"signal: Got signal %u from pid=%u uid=%u: cleaning up",
		signum, info->si_pid, info->si_uid);
#else
void deadly_signal(int signum)
{
	mesg(M_INFO,"signal: Got signal %u, cleaning up", signum);
#endif
	fflush(stdout);
	capture_stop();
}

#ifdef HAVE_SIGACTION 
void hup_signal(int signum, siginfo_t *info, void *foo)
{
	mesg(M_INFO,"signal: Got signal %u from pid=%u uid=%u: rehupping",
		signum, info->si_pid, info->si_uid);
#else
void hup_signal(int signum)
{
	mesg(M_INFO,"signal: Got signal %u, rehupping", signum);
#endif
	fflush(stdout);
	capture_interrupt();
}

int main ( int argc, char **argv ) 
{
#ifdef HAVE_SIGACTION 
	struct sigaction sig;
#endif
	u_int32_t i;

	pbuflen=pbuflen_orig=8192;
	if ( !(pbuf_orig=pbuf=malloc(8192)) ) {
		perror("malloc()");
		exit(1);
	}

	mesg(M_INFO,"Firestorm v" VERSION);
	mesg(M_INFO,"Copyright (c) 2002 Gianni Tedesco");
	mesg(M_INFO,"This program is released under the terms of the "
		"GNU GPL version 2 (see: COPYING)");

#ifdef HAVE_TZSET
	/* Set timezone from locale */
	tzset();
#endif

	/* Initialise our subsystems */
	for(i=0; subsys_init[i]; i++)
		subsys_init[i]();

	/* Initialise plugin subsystem */
	loader_init(plugin_callback);

	/* Read in all configurations */
	conf_cmdline(argc, argv);
	conf_local_config();
	mesg_swap();
	mesg_mode=1;
	conf_background();
	alert_conf_go();

	mesg(M_INFO,"signature: %u signatures loaded", signature_count);

	/* Catch INT and TERM for clean exits */
	/* Catch HUP to rotate logs */
#ifdef HAVE_SIGACTION 
	sig.sa_sigaction=deadly_signal;
	sig.sa_flags=SA_SIGINFO;
	sigemptyset(&sig.sa_mask);
	sigaddset(&sig.sa_mask, SIGINT);
	sigaddset(&sig.sa_mask, SIGTERM);
	sigaction(SIGINT, &sig, NULL);
	sigaction(SIGTERM, &sig, NULL);

	sig.sa_sigaction=hup_signal;
	sig.sa_flags=SA_SIGINFO;
	sigemptyset(&sig.sa_mask);
	sigaction(SIGHUP, &sig, NULL);

	sig.sa_handler=SIG_IGN;
	sig.sa_flags=SA_SIGINFO;
	sigemptyset(&sig.sa_mask);
	sigaction(SIGPIPE, &sig, NULL);
#else
	signal(SIGINT, deadly_signal);
	signal(SIGTERM, deadly_signal);
	signal(SIGHUP, hup_signal);
	signal(SIGPIPE, SIG_IGN);
#endif

	/* This runs the first capture to completion */
	capture_go();

	cleanup(EXIT_OK, "Firestorm exiting normally");
	exit(0);
}
